/*
    selection range API
    // 1.获取光标结束位置
    var end = textarea.selectionEnd
    // 2.通过匹配光标之前文本中的换行符计算所在行
    var row = textarea.value.substring(0, end).match(/\r\n|\r|\n/).length
    // 3.计算 top，行高 * 行数 + 上填充 + 边框宽度
    var top = lineHeight * (row + 1) + paddingTop + borderWidth
    // 4.获取光标左侧的文本
    var leftText = textarea.value.split(/\r\n|\r|\n/)[row]
    // 5.影响一段文字所占宽度的因素太多，除字体大小、中英文、符号、字符间距等，还有字体、浏览器、系统等客观因素
    // var left = ...
    这个方案的思路是没问题的，但是考虑所有问题的成本太高。
    虽然可以创建测试元素去计算文本宽度，但这个方案本身是从严谨的角度出发的。与其混在一块，直接用取巧的办法更简单。
*/

import $ from '../util/dom-core.js'
import { UA, getParentByClassname, getParentNodeByClass, isEmptyElement } from '../util/util.js'

// 构造函数
function API(editor) {
    this.editor = editor;
    this._currentRange = null;
    this.rangeTemp = null;
}

// 修改原型
API.prototype = {
    constructor: API,
    fillChar: '\u200B',
    // 置空 range 对象
    setEmptyRange: function () {
        this._currentRange = null;
    },
    // 获取 range 对象
    getRange: function () {
        return this._currentRange
    },
    // 获取光标位置
    getCursorPosition: function () {
      if (!this._currentRange) {
        this.saveRange();
      }
      return {
        start: this._currentRange.startOffset,
        end: this._currentRange.endOffset
      }
    },
    // 保存选区
    saveRange: function (_range) {
        const selection = window.getSelection()
        
        if (_range) {
            // 保存已有选区
            this._currentRange = _range;
            return this._currentRange;
        }

        // 获取当前的选区
        if (selection.rangeCount === 0) {
            return false;
        }
        const range = selection.getRangeAt(0)

        // 判断选区内容是否在编辑内容之内
        const $containerElem = this.getSelectionContainerElem(range)
        if (!$containerElem) {
            return false;
        }

        // 判断选区内容是否在不可编辑区域之内
        const editor = this.editor
        const $textElem = editor.page.$el;
        if ($textElem.isContain($containerElem)) {
          // 是编辑内容之内的
          this._currentRange = range;        
        }
    },
    isFillChar: function (node,isInStart) {
      if(node.nodeType != 3) {
        return false;
      }
      let text = node.nodeValue;
      if(isInStart){
          return new RegExp('^' + this.fillChar).test(text)
      }
      return !text.replace(new RegExp(this.fillChar,'g'), '').length
    },
    // 折叠选区
    collapseRange: function (toStart) {
        if (toStart == null) {
            // 默认为 false
            toStart = false
        }
        const range = this._currentRange
        if (range) {
            range.collapse(toStart)
        }
    },
    storeRange: function (range) {
      range = range || this._currentRange;
      this.rangeTemp = {
        start: {
          dom: range.startContainer,
          offset: range.startOffset,
        },
        end: {
          dom: range.endContainer,
          offset: range.endOffset,
        },
        collapsed: range.collapsed
      };
    },
    recoverRange: function (rangeTempArg) {
      let rangeTemp = rangeTempArg || this.rangeTemp;
      if (rangeTemp && rangeTemp.start) {
        const selection = window.getSelection()
        let startDom = rangeTemp.start.dom;
        let column = getParentNodeByClass(startDom, 'js-column');
        if (column) {     
          let $startDom = $(
              rangeTemp.start.dom.nodeType === 3 ? rangeTemp.start.dom.parentNode : rangeTemp.start.dom
          ); 
          this.createRangeByElem($startDom, true);
          this.saveRange();          
          let range = this._currentRange;
          try {
            range.setStart(rangeTemp.start.dom, rangeTemp.start.offset);
            range.setEnd(rangeTemp.end.dom, rangeTemp.end.offset);
            this.saveRange(range);
            rangeTemp.collapsed && this.collapseRange();
          } catch (error) {
          }
        }
      }
    },
    // 选中区域的文字
    getSelectionText: function () {
        const range = this._currentRange
        if (range) {
            return this._currentRange.toString()
        } else {
            return ''
        }
    },
    // 选区的 $Elem
    getSelectionContainerElem: function (range) {
        range = range || this._currentRange
        let elem
        if (range) {
            elem = range.commonAncestorContainer
            return $(
                elem.nodeType === 1 ? elem : elem.parentNode
            )
        }
    },
    getSelectionStartElem: function (range) {
        range = range || this._currentRange
        let elem
        if (range) {
            elem = range.startContainer
            return $(
                elem.nodeType === 1 ? elem : elem.parentNode
            )
        }
    },
    getSelectionEndElem: function (range) {
        range = range || this._currentRange
        let elem
        if (range) {
            elem = range.endContainer
            return $(
                elem.nodeType === 1 ? elem : elem.parentNode
            )
        }
    },

    // 选区是否为空
    isSelectionEmpty: function () {
        const range = this._currentRange
        if (range && range.startContainer) {
            if (range.startContainer === range.endContainer) {
                if (range.startOffset === range.endOffset) {
                    return true
                }
            }
        }
        return false
    },

    // 恢复选区
    restoreSelection: function () {
        if (!this._currentRange) {
          return '';
        }
        const selection = window.getSelection()
        selection.removeAllRanges()
        selection.addRange(this._currentRange)
    },

    // 创建一个空白（即 &#8203 字符）选区
    createEmptyRange: function () {
        const editor = this.editor
        const range = this.getRange()
        let $elem

        if (!range) {
            // 当前无 range
            return
        }
        if (!this.isSelectionEmpty()) {
            // 当前选区必须没有内容才可以
            return
        }

        try {
            // 目前只支持 webkit 内核
            if (UA.isWebkit()) {
                // 插入 &#8203
                editor.cmd.do('insertHTML', '&#8203;')
                // 修改 offset 位置
                range.setEnd(range.endContainer, range.endOffset + 1)
                // 存储
                this.saveRange(range)
            } else {
                $elem = $('<strong>&#8203;</strong>')
                editor.cmd.do('insertElem', $elem)
                this.createRangeByElem($elem, true)
            }
        } catch (ex) {
            // 部分情况下会报错，兼容一下
        }
    },

    // 根据 $Elem 设置选区
    createRangeByElem: function ($elem, toStart, isContent) {
        // $elem - 经过封装的 elem
        // toStart - true 开始位置，false 结束位置
        // isContent - 是否选中Elem的内容
        if (!$elem.length) {
            return
        }

        const elem = $elem[0]
        const range = document.createRange()

        if (isContent) {
            range.selectNodeContents(elem)
        } else {
            range.selectNode(elem)
        }

        if (typeof toStart === 'boolean') {
            range.collapse(toStart)
        }
        // 存储 range
        this.saveRange(range)
    },    
    // 光标是否在最后
    isCartEnd: function ($line, range) {
      range = range || this._currentRange;
      if (isEmptyElement($line[0])) {
        return true;
      }
      let lastChild = $line[0].lastChild;      
      if (lastChild.nodeType === 3) {
        if (range.startOffset === range.endOffset && range.startOffset === lastChild.length) {
          return true
        }
      } else {
        return this.isEndDom(range, lastChild);
      }
      return false;
    },
    // 递归判断光标是否在最后
    isEndDom (range, dom) {
      let lastChild = dom.lastChild;
      if (isEmptyElement(dom)) {
        return true;
      }
      if (lastChild.nodeType === 3) {
        if (range.startOffset === range.endOffset && range.startOffset === lastChild.length) {
            return true;
        } else {
            return false;
        }
      } else {
        return isEndDom(range, lastChild);
      }
    }

}

export default API